/*
 * global.c     -global functions
 * distributed as part of _ s u l a _
 * 
 * Author: Tano Fotang, 1997
 * 
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation; either version 2 of the License , or (at your option) 
 * any later version.
 * see the file COPYING for further information.
 *
 */

#include <forms.h>
#include <errno.h>
#include <sys/time.h>
#include <stdarg.h>
#include <ctype.h>
#include <assert.h>
#include "config.h"
#include "aux.h"
#include "sula.h"
#include "chanwin.h"
#include "global.h"



int     atclose_dec_count(FL_FORM * form, void *count)
{
	(*((int *) count))--;
	return FL_OK;
}
int     atclose_reset(FL_FORM * form, void *count)
{
	*((int *) count) = 0;
	return FL_OK;
}

int     atclose_do_close(FL_FORM * form, void *fdui)
{

	return FL_OK;
}

int     dont_close_win(FL_FORM * form, void *fdui)
{

	return FL_IGNORE;
}

int     set_cursor(FL_OBJECT * ob, int event,
		   FL_Coord mx, FL_Coord my, int key, void *xev)
{
	switch (event) {
	    case FL_ENTER:
		    fl_set_cursor(FL_ObjWin(ob), ob->u_ldata);
		    return !FL_PREEMPT;
		    break;
	    case FL_LEAVE:
		    fl_reset_cursor(FL_ObjWin(ob));
		    return !FL_PREEMPT;
		    break;
	    default:
		    return !FL_PREEMPT;
	}
}
void say2(long to, int smaller_window,const char *fmt,...){
#define LINE_SIZE	2048
        char buf[LINE_SIZE];
        va_list ap;
        va_start(ap, fmt);
	// note: vsnprintf is vsprintf without LINE_SIZE. see vsnprint(3)
        if(vsnprintf(buf,LINE_SIZE, fmt, ap)==-1)
	  buf[LINE_SIZE-1]=0;
        va_end(ap);
	say(buf,to,smaller_window);
}

void say (const char *str, long to, int smaller_window){

	if(!win_invalid(to)){
	   if(smaller_window) fl_addto_browser(winstruct[to].chanwin->brow2, str);
	   else fl_addto_browser(winstruct[to].chanwin->browser, str);
	}
	else
	if(main_form)
	    fl_addto_browser(main_form->browser, str);
	else fprintf(stderr,"%s\n",str);
	if(sflags&LOG)
	   if(logFp)
	     if( fprintf(logFp,"%s\n",str)<1){
	      fclose(logFp);
	      logFp=NULL;
	      sflags&= ~LOG;
	      say2(to,1,"write error to %s",log_file);
	     }
}

void error(int code, const char *fmt,...){
/* dsiplay allkinds of messages */
        char buf[MAXLINE];
	int err=errno;
        va_list ap;
        va_start(ap, fmt);
        if(vsnprintf(buf,MAXLINE, fmt, ap)==-1)
	 buf[MAXLINE-1]=0;
        va_end(ap);
	say(buf,-1,1); // logged to logFd too
        if(code<0){
	    sprintf(buf,": %s\n", strerror(err));
	    if(code==-1) say(buf,-1,1);
	    else{
	      fputs(buf,stderr);
	      if(code==-2) exit(1);
	    }
	}else
	if(code>0){
	 if(code==2) raise(SIGSEGV);
	 exit(-1);
	 }
}

int     _save_history(History *h,const char *f)
{
	History *current;
	FILE   *fp;

	if ((fp = fopen(f, "w")) == NULL)
		return errno;

	current = h->right;
	while (current != h) {
		fprintf(fp, "%s\n", current->line);
		current = current->right;
	}
	fclose(fp);
	return 0;
}
int     load_history(const char *f)
{
	FILE   *fp;
	char    s[512 + 1];

	if ((fp = fopen(f, "r")) == NULL)
		return errno;
	while (fgets(s, 512, fp)) {
		s[strlen(s) - 1] = '\0';
		if (s[0])
			AddToHistory(s);
	}
	fclose(fp);
	return 0;
}

extern void addto_cmd_sent(Winstruct * win, const char *cmd, const char *arg)
{
	CmdSent *p = cmdsentStart;

	assert (cmd && win);
	cmdsentStart = (CmdSent *) malloc(sizeof(CmdSent));
	if (cmdsentStart == NULL) {
		return;
	}
	cmdsentStart->cmd = strdup(cmd);
	cmdsentStart->arg = strdup2(arg);
	cmdsentStart->win = win;

	cmdsentStart->next = p;
	return;
}


Winstruct *
        search_winlistbyfd(int fd, int *n)
{
/* 
 * look for the first window with this fd *n is total number of windows found 
 * for this server
 * 
 */
 	int i;
	Winstruct *win=NULL;

	*n = 0;
	for(i=0;i<win_count;i++)
		if (winstruct[i].chanwin && winstruct[i].server && winstruct[i].server->fd == fd) {
			(*n)++;
			if (!win) win=&winstruct[i];
		}
	return win;
}

Winstruct *
        search_winlist_byservname(char *name, int port, int *total)
{
/* 
 * return first window on this server 
 * 
 */
	Winstruct *p = NULL;
	register i;
	*total = 0;

	for(i=0;i<win_count;i++)
		if (winstruct[i].chanwin && winstruct[i].server && 
		!strcasecmp(name, winstruct[i].server->servname) &&
		    port == winstruct[i].server->port) {
			if (!p) p=&winstruct[i];
			(*total)++;
		}
	return p;
}


extern void fit_object_label(FL_OBJECT * ob, const char *label)
{

	if(label==NULL || *label==0) fl_set_object_label(ob,"");
	else{
	int     ll;
	size_t  l;
	char   *buf;

	 buf=(char *)label;
	l = strlen(buf);
	ll = ob->w - 2;
	while (*buf && fl_get_string_width(ob->lstyle, ob->lsize, buf, l) > ll)
		*(buf + --l) = '\0';
	fl_set_object_label(ob, buf);
	}

}


static void free_win_list(void){
/* pop all off*/	
	WinList *p;
	while(servWin){
		p=servWin;
		servWin=servWin->next;
		free(p);
	}
}
int     win_add(Winstruct * win)
{
	WinList *p = servWin;

	servWin = (WinList *) malloc(sizeof(WinList));
	if (!servWin)
		return errno;
	servWin->win = win;
	servWin->next=p;
	return 0;
}

void   *
        build_winlistbyfd(int fd)
{
	register i;
	free_win_list();
	for(i=0;i<win_count;i++)
		if (winstruct[i].chanwin && winstruct[i].server && 
		    fd == winstruct[i].server->fd)
			if(win_add(&winstruct[i])) return NULL;
	return (void *) servWin;
}

void   *
        build_winlistbyservname(char *name, int port)
{
/* 
 * similar to build_winlistbyfd  */

	register i;
	free_win_list();
	for(i=0;i<win_count;i++)
	if (winstruct[i].chanwin && winstruct[i].server && 
	!strcasecmp(name, winstruct[i].server->servname) &&
	    port == winstruct[i].server->port) 
		if(win_add(&winstruct[i])) return (void *)0;
	return (void *) servWin;
}

