/*
   our_guile.c - Scheme support for _sula_

 */
#include <forms.h>
#include "config.h"
#include "aux.h"
#include "sula.h"
#include "chanwin.h"
#include "global.h"
#include "script_gui.h"
#include "script.h"
#include "cmd.h"
#include "scheme.h"
#include "hooks.h"
#include "timer.h"
#include "nicks.h"
#include <arpa/inet.h>  //inet_ntoa

#define MY_NULL gh_list(SCM_UNDEFINED)
#define VAR_STEM_LEN 16
#define MAX_FUNC_NAME_LEN  (VAR_STEM_LEN+12)

#define I_AM_DUMB  0
static char var_stem[VAR_STEM_LEN];  //variable names
#if I_AM_DUMP
static char var_tmp[VAR_STEM_LEN];
#endif
extern get_set_variable(const char *);
extern SCM cmd_set_var(SCM,SCM);
extern FD_scriptcontrol *script_form;
extern int ignored_p(const char *mask, const char *what);

#define SCM_FUNC_ALLOC 20
static int scm_func_allocate(void){
	register i;
	if(scm_func==NULL)
		scm_func=malloc(sizeof(ScmFunc)*SCM_FUNC_ALLOC);
	else
		scm_func=realloc(scm_func,
		   (scm_func_count+SCM_FUNC_ALLOC)*sizeof(ScmFunc));
	if(scm_func==NULL) return (-1);
	for (i=scm_func_count;i< scm_func_count+SCM_FUNC_ALLOC; i++)
		scm_func[i].idx=-1;
	scm_func_count+=SCM_FUNC_ALLOC;
	return (0);
}
extern int define_func(SCM proc){
  register i;
  if(!scm_func && scm_func_allocate()) return -1;
  do{
	for(i=0;i<scm_func_count;i++)
	    if(scm_func[i].idx==-1){
	       char buf[MAX_FUNC_NAME_LEN];
	       
	       scm_func[i].idx=i;
	       assert(gh_procedure_p(proc));
	       if(snprintf(buf,MAX_FUNC_NAME_LEN,"%s-%d",var_stem,i)==-1)
	           buf[MAX_FUNC_NAME_LEN-1]=0;
	       scm_func[i].func=gh_define(buf,proc);
	       return i;
	   }
	if(scm_func_allocate()) return -1;
   }while(1);
}

static SCM s_ignored_p(SCM who, SCM what)
{
 int n;
 char *pat;
 char *mask=gh_scm2newstr(who,&n);
 if(n==0) return SCM_BOOL_F;
 *(mask+n)=0;
 pat=gh_symbol2newstr(what,&n);
 gh_defer_ints();
 n=ignored_p(mask,pat);
 gh_allow_ints();
 return gh_bool2scm(n);
 
}
static int get_valid_window(void)
/* return index of a valid channel window */
{
    register int i;
    Winstruct *win=NULL;

    for (i = 0; i < win_count; i++)
	if (winstruct[i].chanwin){
	    if(!win) win=&winstruct[i];
	    if(winstruct[i].server){
	        if(!winstruct[i].server) win=&winstruct[i];
		if(winstruct[i].server->fd > -1) return i;
	    }
	}
    return win?win-winstruct:-1;
}

static SCM s_get_server_by_win_nr(args)
/* return server connected to a window */
     SCM args;
{
    register int w;

    w = -1;

    if (gh_null_p(args))
	w = get_valid_window();
    else
	w = (int) gh_scm2long(gh_car(args));
    if (win_invalid(w))
	return MY_NULL;
    else {
	Server *s = winstruct[w].server;
	if(s==NULL) return MY_NULL;
	return gh_list(gh_long2scm(s->fd),	/* fd */
		       gh_str02scm(s->servname), /* server "alias" */
		       gh_long2scm(s->port),	/* port */
		       gh_str02scm(s->nick),	/* nick */
		       gh_str02scm(s->ircname),		/* realname */
		       gh_str02scm(s->mode),
		       gh_long2scm(w),
		       gh_str02scm(s->name),	/*server*/
		       gh_str02scm(s->lastInvite),
		       gh_bool2scm((s->flag&SET_AWAY)!=0),
		       gh_long2scm(s->read),	/* bytes read from serevr*/
		       gh_long2scm(s->wrote),	/* bytes written to serevr*/
		       gh_long2scm(s->connect_time),
		       gh_long2scm(s->iNick),
		       SCM_UNDEFINED);
    }

}

static SCM s_get_channel_list_by_win(args)
/* a list of all channels of this window
 */
     SCM args;
{
    int w = -1;
    SCM list=gh_list(SCM_UNDEFINED);
    if (gh_null_p(args))
	w = get_valid_window();
    else
	w = (int) gh_scm2long(gh_car(args));
    if (!win_invalid(w)) {
	Channel *current;
	Winstruct *win = &winstruct[w];
	
	current = win->chanStart->left;
	while (current != win->chanStart) {
	    list = gh_cons(gh_list(gh_long2scm(w),
				   gh_str02scm(current->name),
				   gh_str02scm(current->modes),		/* o,v,n,t etc */
				   gh_long2scm(current->limit),
				   gh_str02scm(current->key),
				   gh_str02scm(current->topic),
				   SCM_UNDEFINED),
			   list);
	    current = current->left;
	}
    }
    return list;
}
static SCM s_get_dcc_list_by_win(args)
/* a list of all DCC chats on this window
 */
     SCM args;
{
    int w = -1;
    SCM list;
    if (gh_null_p(args))
	w = get_valid_window();
    else
	w = (int) gh_scm2long(gh_car(args));
    list = gh_list(SCM_UNDEFINED);
    if (!win_invalid(w)) {
	DCCChat *current;
	Winstruct *win = &winstruct[w];
	
	current = win->dccStart->left;
	while (current != win->dccStart) {
	    list = gh_cons(gh_list(gh_long2scm(w),
	    			   gh_long2scm(current->fd),
				   gh_str02scm(current->nick),
				   gh_str02scm(current->email),
				   gh_str02scm(current->rem_host),
				   gh_long2scm(current->read),
				   gh_long2scm(current->wrote),
				   SCM_UNDEFINED),
			   list);
	    current = current->left;
	}
    }
    return list; 
}
static SCM s_get_current_dcc_nick(win)
     SCM win;
{
    register int w;

    if (gh_null_p(win))
	w = get_valid_window();
    else
	w = (int) gh_scm2long(gh_car(win));
    if (win_invalid(w))
	return SCM_BOOL_F;
    if (winstruct[w].dccCur == winstruct[w].dccStart)
	return SCM_BOOL_F;
    else
	return gh_str02scm(winstruct[w].dccCur->nick);

}
static SCM s_get_query_nick(win)
     SCM win;
{
    register int w;

    if (gh_null_p(win))
	w = get_valid_window();
    else
	w = (int) gh_scm2long(gh_car(win));
    if (win_invalid(w))
	return SCM_BOOL_F;
    return gh_str02scm(winstruct[w].query_nick);
}

static SCM get_channel_info(int w, const char *chan)
{
    if (!win_invalid(w)) {
	Channel *current;
	Winstruct *win = &winstruct[w];

	current = win->chanStart->left;
	while (current != win->chanStart) {
	    if (!strcasecmp(current->name, chan)) {
		SCM list = gh_list(gh_long2scm(w),
				   gh_str02scm(current->name),
				   gh_str02scm(current->modes),		/* o,v,n,t etc */
				   gh_long2scm(current->limit),
				   gh_str02scm(current->key),
				   gh_str02scm(current->topic),
				   SCM_UNDEFINED);

		return list;
	    }
	    current = current->left;
	}
    }
    return gh_list(SCM_UNDEFINED);
}
static SCM s_get_channel_info(args)
/* info about a channel
   (window  channel modes  topic  etc) */
     SCM args;
{
    int w;
    char *ch = 0;
    SCM res;

    if (!gh_null_p(args)) {
	w = 0;
	ch = gh_scm2newstr(gh_car(args), &w);
	if (w > 0)
	    *(ch + w) = 0;
	w = -1;
	args = gh_cdr(args);
	/*if(win_invalid(w)) return MY_NULL; */
	if (!gh_null_p(args))
	    w = (int) gh_scm2long(gh_car(args));
    }
    else
	w = -1;
    if (win_invalid(w)) {
	w = get_valid_window();
	if (win_invalid(w))
	    return MY_NULL;
    }
    if (winstruct[w].chanCur == winstruct[w].chanStart)
	return MY_NULL;
    if (ch == 0)
	ch = strdup(winstruct[w].chanCur->name);
    gh_defer_ints();
    res = get_channel_info(w, ch);
    free(ch);
    gh_allow_ints();
    return res;

}

static SCM s_win_msg(msg, opt)
     SCM msg;
     SCM opt;
{
    char *str;
    int l;

    str = gh_scm2newstr(msg, &l);
    *(str + l) = 0;
    if (!gh_null_p(opt))
	l = (int) gh_scm2long(gh_car(opt));
    else
	l = get_valid_window();
    if (win_invalid(l))
	return SCM_BOOL_F;
    gh_defer_ints();
    fit_object_label(winstruct[l].chanwin->WinStatusLine, str);
    XFlush(fl_display);
    gh_allow_ints();
    free(str);
    return (SCM_BOOL_T);
}
#if 0
static SCM s_win_status(msg, opt)
     SCM msg;
     SCM opt;
{
    char *str;
    int l;

    str = gh_scm2newstr(msg, &l);
    *(str + l) = 0;
    if (!gh_null_p(opt))
	l = (int) gh_scm2long(gh_car(opt));
    else
	l = get_valid_window();
    if (win_invalid(l))
	return SCM_BOOL_F;
    gh_defer_ints();
    fit_object_label(winstruct[l].chanwin->...., str);
    XFlush(fl_display);
    gh_allow_ints();
    free(str);
    return (SCM_BOOL_T);
}
#endif
static SCM get_server_channels(const char *servername, int port)
{
    SCM list;
    gChannel *p;

    list = gh_list(SCM_UNDEFINED);
    p = chanStart;
    while (p != chanEnd) {
	if (servername == NULL || !strcasecmp(p->server, servername)) {
	    if (port == p->port || port == 0)
		list = gh_cons(gh_list(gh_str02scm(p->name),
				       gh_str02scm(p->server),
				       gh_long2scm(p->port),
				       SCM_UNDEFINED
			       ), list);
	}
	p = p->next;
    }
    return list;

}

static SCM s_get_channel_list_by_server(args)
/* 
   (gs-server->channels [server [port]])
   servername=NULL --> all servers 
   port ==0 any ports
   port<0 -->default port (usulally 6667) 
   return lists of (channel server  port ) */
     SCM args;
{
    char *servername = NULL;
    int port = 0;
    SCM list;

    if (!gh_null_p(args)) {
	int l = 0;

	if (gh_string_p(gh_car(args))) {
	    servername = gh_scm2newstr(gh_car(args), &l);
	    if (l < 1)
		servername = NULL;
	    else
		*(servername + l) = 0;
	    args = gh_cdr(args);
	}
	if (!gh_null_p(args) && gh_exact_p(args))
	    port = (int) gh_scm2long(gh_car(args));
    }
    if (port < 0)
	port = DEFAULT_PORT;
    gh_defer_ints();
    list = get_server_channels(servername, port);
    gh_allow_ints();
    if (servername)
	free(servername);
    return list;
}

static SCM chan_users(int w, Channel * c)
{
/* (win  channel  nick  mode email_addr  ircname) */
    SCM list = gh_list(SCM_UNDEFINED);
    User *u;

    u = c->userStart;
    while (u != c->userEnd) {
	list = gh_cons(gh_list(gh_long2scm(w),
			       gh_str02scm(c->name),
			       gh_str02scm(u->name),
			       gh_str02scm(u->mode),
			       gh_str02scm(u->address),
			       gh_str02scm(u->ircname),
			       SCM_UNDEFINED
		       ),
		       list);
	u = u->next;
    }
    return list;
}

static SCM s_get_nicklist_by_channel(args)
/* (win  channel  nick  mode email_addr  ircname) */
/* (dsdsdas channelname window_nr) */
     SCM args;
{
    char *ch = 0;
    SCM list;

    list = gh_list(SCM_UNDEFINED);
    if (!gh_null_p(args)) {
	int l = 0;
	SCM tmp;

	ch = gh_scm2newstr(gh_car(args), &l);
	if (l < 1)
	    return list;
	*(ch + l) = 0;
	tmp = gh_cdr(args);
	l = -1;
	if (!gh_null_p(tmp))
	    l = (int) gh_scm2long(gh_car(tmp));
	else
	    l = get_valid_window();
	if (l > -1 && l < win_count) {
	    Channel *current;
	    Winstruct *win = &winstruct[l];

	    current = win->chanStart->left;
	    while (current != win->chanStart)
		if (!strcasecmp(ch, current->name)) {
		    list = chan_users(l, current);
		    free(ch);
		    break;
		}
		else
		    current = current->left;
	}
    }
    return list;
}
static SCM cmd_last_part(win)
     SCM win;
{
    register int w;

    if (gh_null_p(win))
	w = get_valid_window();
    else
	w = (int) gh_scm2long(gh_car(win));
    if (win_invalid(w))
	return MY_NULL;
    return gh_list(
         gh_str02scm(winstruct[w].lastPart->name),
	 gh_str02scm(winstruct[w].lastPart->key),
	 SCM_UNDEFINED);
}
static SCM cmd_last_join(win)
     SCM win;
{
    register int w;

    if (gh_null_p(win))
	w = get_valid_window();
    else
	w = (int) gh_scm2long(gh_car(win));
    if (win_invalid(w))
	return MY_NULL;
    return gh_list(
         gh_str02scm(winstruct[w].lastJoin->name),
	 gh_str02scm(winstruct[w].lastJoin->key),
	 SCM_UNDEFINED);
}
static SCM s_get_current_by_channel(win)
     SCM win;
{
    register int w;

    if (gh_null_p(win))
	w = get_valid_window();
    else
	w = (int) gh_scm2long(gh_car(win));
    if (win_invalid(w))
	return SCM_BOOL_F;
    if (winstruct[w].chanCur == winstruct[w].chanStart)
	return SCM_BOOL_F;
    else
	return gh_str02scm(winstruct[w].chanCur->name);

}
//
// nick groups
//
static SCM s_list_nick_groups_cmd(void){
//( i . label )
    register int n;
    SCM result = gh_list(SCM_UNDEFINED);
    for(n=0;n<iNickListCount;n++)
       if(aNickList[n].label)
            result = gh_cons(gh_list(gh_long2scm(n),
			       gh_str02scm(aNickList[n].label),
			       SCM_UNDEFINED
		       ),
		       result);
   return result;
}
static SCM s_create_nickgroup_cmd(SCM name)
{
 int n;
 char *label=gh_scm2newstr(name,&n);

 if(n==0) return SCM_BOOL_F;
 gh_defer_ints();
 n=new_nicklist(label);
 gh_allow_ints();
 //if(n==-1)  scm_syserror("s_create_nickgroup_cmd",strerror(errno),
 free(label);
 return gh_int2scm(n);
}
static SCM s_addto_nickgroup_cmd(SCM group, SCM val)
{
  int  n;
  char *nick;
  
  nick=gh_scm2newstr(val,&n);
  if(n==0) return SCM_BOOL_F;
  n=gh_scm2int(group);
  if(n<0)
     return SCM_BOOL_F;// scm_wrong_type_arg("s_addto_nickgroup",1,group);
  gh_defer_ints();
  aNickList[n].current=add_nick_to_list(n,nick);
  gh_allow_ints();
  free(nick);
  return SCM_BOOL_T;
}
static SCM s_list_group_nicks_cmd(SCM group)
{
  SCM result = gh_list(SCM_UNDEFINED);
  int n=gh_scm2int(group);
  if(n<0) return result; //scm_wrong_type_arg("s_list_group_nicks",1,group);
  if(n<iNickListCount && aNickList[n].label){
       Nicks *p=aNickList[n].start->left;
       while(p!=aNickList[n].start){
	   result=gh_cons(gh_str02scm(p->nick), result);
	   p=p->left;
	}
  }
  return result;
}
static SCM s_set_server_nick_group(SCM server, SCM val)
{
  int i,n;
  Server *s;
  char *name;
  u_short port;
  if(gh_null_p(server)) return SCM_BOOL_F;
  name=gh_scm2newstr(gh_cadr(server),&i);
  if(i==0) return SCM_BOOL_F;
  port=gh_scm2ulong(gh_caddr(server));
  gh_defer_ints();
  s=search_server_tree_byname(name,port);
  gh_allow_ints();
  free(name);
  if(s==NULL) return SCM_BOOL_F;
  i=gh_scm2int(val);
  if(i<0 || i>iNickListCount || aNickList[n].label==NULL)
    return SCM_BOOL_F;//scm_wrong_type_arg("s_set_server_nick_group",2,val);
  s->iNick=i;
  name=gh_scm2newstr( gh_list_ref(server,gh_int2scm(3)),&n);
  	//gh_cadddr(server),&n);nick
  if(n>0){
	 add_nick_to_list(s->iNick, name);
	 free(name);
  }
  return SCM_BOOL_T;
}
//----------EO nick ggroups
static SCM s_win_list(void)
{
    register int i;
    SCM result = gh_list(SCM_UNDEFINED);

    for (i = 0; i < win_count; i++)
	if (winstruct[i].chanwin)
	    result = gh_cons(gh_long2scm(i), result);
    return result;
}
static SCM s_win_create(void)
{
    if(main_form){
    gh_defer_ints();
    fl_trigger_object(main_form->newwin);
    gh_allow_ints();
    }
    return SCM_UNDEFINED;
}

static SCM s_win_destroy(win)
     SCM win;
{
    int w = -1;

    w = (int) gh_scm2long(win);
    if (win_invalid(w))
	return SCM_BOOL_F;
    else {
	gh_defer_ints();
	del_chanwin_from_list(&winstruct[w]);
	gh_allow_ints();
    }
    return SCM_BOOL_T;

}
static SCM s_get_string(args)
     SCM args;
{
    char *prompt, *def_val = NULL, *result = 0;
    int len;
    SCM ret;

    if (gh_null_p(args)) {
	prompt = strdup("Input");
	def_val = NULL;
    }
    else {
	prompt = gh_scm2newstr(gh_car(args), &len);
	if (len < 1)
	    return MY_NULL;
	*(prompt + len) = 0;
	args = gh_cdr(args);
	if (!gh_null_p(args)) {
	    def_val = gh_scm2newstr(gh_car(args), &len);
	    if (len > 0)
		*(def_val + len) = 0;
	}
    }
    gh_defer_ints();
    result = getstr2(prompt,
		     sula_NAME sula_VERSION ": Get a string",
		     def_val,
		     0);
    gh_allow_ints();
    free(prompt);
    free(def_val);
    ret = gh_str02scm(result);
    free(result);
    return ret;
}

static SCM s_next_string_cmd(str,pos)
     SCM str;
     SCM pos;
{
    char *s,*res;
    int l;
    SCM r=SCM_BOOL_F;

    s = gh_scm2newstr(str, &l);
    if (l==0)
	return SCM_BOOL_F;
    *(s + l) = 0;
    l=gh_scm2long(pos);
    if(l<0){
          scm_wrong_type_arg("s_next_string_cmd",2,pos);
	  return r;
       }
    gh_defer_ints();
    res=nextstr(s,l);
    gh_allow_ints();
    if(*res) r= gh_str02scm(res);
    free(res);
    return r;
}

//////////////////////////////////
// Named connections		//
//////////////////////////////////

static SCM s_list_nc_cmd(void)
// return a list of all NC indices
{
 register i;
 SCM result = gh_list(SCM_UNDEFINED);
 for(i=0;i<sp_count;i++)
   if(s_pipe[i].fd!=-1)
	    result = gh_cons(gh_long2scm(i), result);
    return result;
}
static SCM s_nc_cmd(SCM n)
// take an index into NC list and return info
{
      int nc=(int) gh_scm2ulong(n);
      if(nc<0|| nc>sp_count || s_pipe[nc].fd==-1)
        return MY_NULL;
    else {

	S_Pipe *s = &s_pipe[nc];
	SCM type, list;
	switch(s->type){
	  case CON_SCRIPT:
	     type= gh_symbol2scm("exec");
	     list=gh_list(gh_int2scm(s->to.proc->pid),
	                gh_str02scm(s->to.proc->sys_cmd),
			 SCM_UNDEFINED);
		break;
	  case CON_SERVER:
	       type= gh_symbol2scm("tcp/ip");
	       list=gh_list(
	                gh_str02scm(s->to.server.name),
			gh_int2scm(s->to.server.port),
			gh_str02scm(inet_ntoa(s->to.server.sin_addr)),
			gh_str02scm(s->to.server.args),
			 SCM_UNDEFINED);
	       break;
	  default:
	     type= gh_symbol2scm("unknown");
	     list=MY_NULL;
	     break;
	   }
	  return gh_list(gh_int2scm(nc),
		       gh_int2scm(s->fd),	/* fd */
		       gh_str02scm(s->name), /* name of connection*/
		       type,
		       list,
		       SCM_UNDEFINED);
    }
}
static SCM s_connect_cmd(SCM name,SCM server,SCM port,SCM args)
{
    int len;
    char  *bez;
    char *to,*misc=NULL;
    u_short pport;
    S_Pipe *s=NULL;

    to = gh_scm2newstr(server, &len);
    if (len ==0)
	return SCM_BOOL_F;
    *(to + len) = 0;
    bez = gh_scm2newstr(name, &len);
    if (len ==0)
	return SCM_BOOL_F;
    *(bez + len) = 0;
    pport=gh_scm2ulong(port);
    if(gh_null_p(args))
      misc=NULL;
    else
      misc=gh_scm2newstr(args,&len);
    gh_defer_ints();
    s=make_connection(bez,to,pport,misc);
    gh_allow_ints();
    free(to);
    free(bez);
    if(misc) free(misc);
    return s?gh_int2scm(s-s_pipe):SCM_BOOL_F;
}
static SCM s_run_program_cmd(name, args)
     SCM args;
     SCM name;
{
    int len;
    char *prog, *bez;

    prog = gh_scm2newstr(args, &len);
    if (len < 1)
	return SCM_BOOL_F;
    *(prog + len) = 0;
    bez = gh_scm2newstr(name, &len);
    if (len < 1)
	return SCM_BOOL_F;
    *(bez + len) = 0;
    gh_defer_ints();
    launch_prog(bez, prog);
    gh_allow_ints();
    free(prog);
    free(bez);
    return SCM_BOOL_T;
}
////////////////////////////////////////////
static SCM s_exit_cmd(exit_code)
     SCM exit_code;
{
    int c = (int) gh_scm2long(exit_code);

    gh_defer_ints();
    fl_finish();
    exit(c);
    gh_allow_ints();
    return SCM_UNDEFINED;
}
static SCM s_echo_cmd(str, args)
/* echo msg [browser [window]]
   (gs-echo "blabla" 2) */
     SCM str;
     SCM args;
{
    int w, b = 0;
    char *s;

    s = gh_scm2newstr(str, &w);
    *(s + w) = 0;
    w = -1;
    if (!gh_null_p(args)) {
	w = (int) gh_scm2long(gh_car(args));
	args = gh_cdr(args);
	if (!gh_null_p(args))
	    b = (int) gh_scm2long(gh_car(args));
    }
    gh_defer_ints();
    say(s, w, b);
    gh_allow_ints();
    return SCM_UNDEFINED /*SCM_BOOL_F */ ;
}

static SCM s_exec_cmd(cmd, opt)
     SCM cmd;
     SCM opt;
{
    char *scmd;
    int l;

    scmd = gh_scm2newstr(cmd, &l);
    if (l < 1)
	return SCM_BOOL_F;
    *(scmd + l) = 0;
    if (!gh_null_p(opt))
	l = (int) gh_scm2long(gh_car(opt));
    else
	l = get_valid_window();
    gh_defer_ints();
    l = new_process_cmd(scmd, l);
    gh_allow_ints();
    return (gh_bool2scm(0 == l));
}

// H o o k s

static SCM hook_cmds(SCM args,SCM arg1,SCM arg2,SCM rest, char preempt)
{
    int n;
    char *pat = NULL;
    MsgNr type = INVALID;
    int numeric = 0;
    int ser_no = 0;
    SCM proc;
    On  w;
    unsigned char pos=1;
    if (gh_string_p(args)) {
	char *cmd;
	pos=2;
	cmd = gh_scm2newstr(args, &n);
	if (n < 1)
	    scm_wrong_type_arg("hook_cmds",1,args);
	*(cmd + n) = 0;
	if (*cmd == '#') {
	    strshift(cmd,1);
	    if (*cmd == '\0')
		scm_wrong_type_arg("hook_cmds",1,args);
	    if (!gh_exact_p(arg1))
	        scm_wrong_type_arg("hook_cmds",2,arg1);    
	    ser_no = (int) gh_scm2long(arg1);
	    pos++;
	};
	numeric = strtol(cmd, &pat, 0);
	if (*pat) {
	    type = get_type(cmd);
	    numeric = 0;
	}
	free(cmd);
    }
    else {
	if (!gh_exact_p(args))
	 scm_wrong_type_arg("hook_cmds",1,args);
	numeric = (int) gh_scm2long(args);
	pos=2;
    }
    if (type == INVALID && numeric < 1)
	scm_wrong_type_arg("hook_cmds",1,args);
    if(pos==2){
       pat = gh_scm2newstr(arg1, &n);
       if(n==0) scm_wrong_type_arg("hook_cmds",pos,arg1);
       *(pat+n)=0;
       proc = arg2;
       if (!gh_procedure_p(proc))
        scm_wrong_type_arg("hook_cmds",pos,proc);
    }else
    {
       pat = gh_scm2newstr(arg2, &n);
       if(n==0) scm_wrong_type_arg("hook_cmds",pos,arg2);
       *(pat+n)=0;
       if(gh_null_p(gh_car(rest))){
             fprintf(stderr,"Missing hook procedure\n");
	     return SCM_BOOL_F;
	}
	proc=gh_car(rest);
	if (!gh_procedure_p(proc))
           scm_wrong_type_arg("hook_cmds",pos,proc);
   }
   w.flag = HK_SCM;
   if(preempt)
      w.flag |= HK_PREEMPTIVE;
    w.pattern = pat;
    w.cmd.scm=-1;
    w.id = -1;			/* used by list_hooks() */
    w.ser_number =ser_no;
    gh_defer_ints();
    if (type != INVALID)
	add_hook2(type, &w,(void *)&proc);
    else
	add_numhook2(numeric, &w,(void *)&proc);
    if (script_form)
	list_hooks(1);
    gh_allow_ints();
    return SCM_BOOL_T;
    gh_allow_ints();
}
static SCM s_on_cmd(SCM args,SCM arg1,SCM arg2,SCM rest)
{
 return(hook_cmds(args,arg1,arg2,rest,0));   
}

static SCM s_preempt_cmd(SCM args,SCM arg1,SCM arg2,SCM rest)
{
    return(hook_cmds(args,arg1,arg2,rest,1));
}

static SCM s_remove_preemptive_cmd(SCM args)
{
int ll;
char *str=gh_scm2newstr(args,&ll);
if(ll==0) return  SCM_BOOL_F;
*(str+ll)=0;
gh_defer_ints();
delete_hook(str,1);
gh_allow_ints();
return SCM_BOOL_T;
}

//////////////////////////
// T i m e r s		//
//////////////////////////
static SCM s_timer_list_cmd(void)
//(ref . alarmORclock?)
{
 register i;
 SCM result = gh_list(SCM_UNDEFINED);

 for(i=0;i<timer_count;i++)
   if(aTimer[i].type!=-1)
       result = gh_cons(gh_list(gh_long2scm(aTimer[i].ref),
                                gh_bool2scm(aTimer[i].flag&TIMER_ALARM),
				SCM_UNDEFINED
			),
			result);
  return result;      
}
static int get_free_ref(){
// the IDs that we automatically assign to timers should be -ve.
// this makes it possible for the user to decide to assign only +ve ID
// when s/he explicitely specifies a timer ID.
// 19Nov98 ft22
#define MAX_REFNUM INT_MIN //preposterous! tell me that...
 register i,j;
 char got_it=1;
 for(j=-1; j>MAX_REFNUM;j--,got_it=1){
   for(i=0;i<timer_count;i++){
      if(aTimer[i].type!=-1)
        if(aTimer[i].ref==j){got_it=0; break;}
   }
   if(got_it) return j;
 }
 // impossible!!
 return INT_MIN;
}


static SCM s_alarm_cmd(args)
  SCM args;
{
/*
 'ref <refnum> <secs> <proc>
 <secs> <proc>
*/
   SCM arg1;
   SCM proc;
   int sec;
  Timer t;
  unsigned char pos=1;

   if (gh_null_p(args))
	return SCM_BOOL_F;
   arg1=gh_car(args);
   args=gh_cdr(args);
   if(gh_symbol_p(arg1)){
	if(gh_equal_p(arg1, gh_symbol2scm("ref"))){
	   if (gh_null_p(args))
	     return SCM_BOOL_F;
	   if (!gh_exact_p(gh_car(args))){
		 scm_wrong_type_arg("s_alarm_cmd",2,gh_car(args));
	         return SCM_BOOL_F;
	   }
	   t.ref=gh_scm2long(gh_car(args));
	   pos+=2;
	   //if(gh_length(args)<2))
	   // scm_wrong_num_args(
	   args=gh_cdr(args); // time proc
	   if (gh_null_p(args)){
	     return SCM_BOOL_F;
	     }
	  arg1=gh_car(args); // seconds
	  args=gh_cdr(args); // proc
      }
      //else scm_wrong_type_arg("s_alarm_cmd",1,arg1);
  }
   else
      	  t.ref=get_free_ref();
   sec=(int)gh_scm2long(arg1);
  if(sec<0){
    scm_wrong_type_arg("s_alarm_cmd",pos,arg1);
     return SCM_BOOL_F;
  }
  pos++;
  if (gh_null_p(args)){
		// scm_wrong_type_arg("s_alarm_cmd",pos,gh_car(args));
	     return SCM_BOOL_F;
	     }
  proc = gh_car(args);
  if (!gh_procedure_p(proc)){
    scm_wrong_type_arg("s_alarm_cmd",pos,proc);
     return SCM_BOOL_F;
  }
  t.type=ALARM_SCM;
  t.flag=TIMER_ALARM;
  t.cmd.scm=define_func(proc);
  if(t.cmd.scm <0) error(-2, "Memory error");
  gh_defer_ints();
  add_timer(&t,(void *)&sec);
  gh_allow_ints();
  return gh_int2scm(t.ref);
}
static SCM s_alarm_del_cmd(sref)
  SCM sref;
{
  int ref;
  ref=(int) gh_scm2long(sref);
  gh_defer_ints();
  del_alarm(ref,ALARM_SCM);
  gh_allow_ints();
  return SCM_BOOL_T;
}
static SCM s_clock_del_cmd(sref)
  SCM sref;
{
  int ref;
  ref=(int) gh_scm2long(sref);
  gh_defer_ints();
  del_clock(ref,ALARM_SCM);
  gh_allow_ints();
  return SCM_BOOL_T;
}

static SCM s_clock_cmd(args)
     SCM args;
{
/*
 clock 'ref 200 '(20 15 30 20 12 1998) proc
 clock '(20 15 30 20 12 1998) proc
 */

   SCM arg1;
    int ref=0;
   
   if (gh_null_p(args))
	goto usage;
   arg1=gh_car(args);
   args=gh_cdr(args);
   if(gh_symbol_p(arg1)){
	if(gh_equal_p(arg1, gh_symbol2scm("ref"))){
	   if (gh_null_p(args) || !gh_exact_p(gh_car(args)))
	         goto usage;
	   ref=gh_scm2long(gh_car(args));
	   args=gh_cdr(args); // proc
	   if (gh_null_p(args))
          goto usage;
	  arg1=gh_car(args); // date
	  args=gh_cdr(args);
       }
   }
   else
      	  ref=get_free_ref();
   if(gh_list_p(arg1) && gh_length(arg1)>0)
   {
       Timer t;
       struct tm *when;
       time_t now;
       int year, mon;
       int i=0;
       int *var[6];
	SCM proc;

       proc = gh_car(args);
       if (!gh_procedure_p(proc))
	goto usage;
  #warning (19Nov98) No plausibilty tests performed on the date.Please implement.
       now=time(NULL);
       when=localtime(&now);
       when->tm_sec=0;
       year=1900+when->tm_year;
       mon=1+when->tm_mon;
       var[0]=&when->tm_hour;
       var[1]=&when->tm_min;
       var[2]=&when->tm_sec;
       var[3]=&when->tm_mday;
       var[4]=&mon;
       var[5]=&year;
       while(!gh_null_p(arg1) && i<6){
          *(var[i])=gh_scm2int(gh_car(arg1));
	  arg1=gh_cdr(arg1);
	  i++;
      }
      if(i>5) year+=1900;
      if(year<when->tm_year+1900){
	  fprintf(stderr, "clock- Invalid year given: %d\n", year);
	  return  SCM_BOOL_F;
     }
     when->tm_year=year-1900;
     // this is lame
     when->tm_mon=(mon-1)%12;
     when->tm_mday%=31;
     when->tm_hour%=24;
     when->tm_sec%=60;

     t.ref=ref;
     t.type=ALARM_SCM;
     t.flag=TIMER_CLOCK;
     gh_defer_ints();
     t.cmd.scm=define_func(proc);
     if(t.cmd.scm <0) error(-2, "Memory error");
     if((i=add_timer(&t,(void *)when))==-1)
           error(-2,"malloc error");
     gh_allow_ints();
     #if DEBUG
     fprintf(stderr,"clock %d set for %02d:%02d:%02d %02d.%02d.%d\n",aTimer[i].ref,
     aTimer[i].expire.date->tm_hour, aTimer[i].expire.date->tm_min,
     aTimer[i].expire.date->tm_sec,
     
     aTimer[i].expire.date->tm_mday,aTimer[i].expire.date->tm_mon+1,
     aTimer[i].expire.date->tm_year+1900);
     #endif
     return (gh_int2scm(ref));
    }
    usage:
    // blabla
    fprintf(stderr,"Usage: (gs-clock ['ref <id>] '(<HH> <MM> <SS> <DD> <mm> <YYYY>) <proc>)\n");
    return SCM_BOOL_F;     
}
//////////////////////////-timers end-----//////////


// these build-in commands may not be replaced or removed
static char *immune[]={"who","mode", NULL};
static SCM s_remove_command_cmd(SCM cmd)
{
 int l;
 char *name;
 char **p=immune;
 Cmd *c;
 name=gh_scm2newstr(cmd, &l);
 if (l ==0)
	return SCM_BOOL_F;
 gh_defer_ints();
 c = find_cmd(name, 1);
 gh_allow_ints();
 if(c) while(*p)
   if(!strcasecmp(*p, c->name)){
       #if DEBUG
       fprintf(stderr, "command %s may not be removed.\n",c->name);
       #endif
       return  SCM_BOOL_F;
     }
     else p++;
 gh_defer_ints();
   remove_cmd(name);
   gh_allow_ints();
 return SCM_BOOL_T;
}
static SCM s_remove_ctcp_command_cmd(SCM cmd)
{
 int l;
 char *name;

 name=gh_scm2newstr(cmd, &l);
 if (l ==0)
	return SCM_BOOL_F;
 gh_defer_ints();
 remove_ctcp_cmd(name);
 gh_allow_ints();
 return SCM_BOOL_T;
}
static SCM s_reg_command(name, descr, func)
/*register a new command */
     SCM name;
     SCM descr;
     SCM func;
{
    Cmd cmd, *c = 0;
    int l;

    cmd.name = gh_scm2newstr(name, &l);
    if (l ==0)
	return SCM_BOOL_F;
    cmd.help = gh_scm2newstr(descr, &l);
    if (l < 1)
	return SCM_BOOL_F;
    cmd.type = CMD_SCM | CMD_NORMAL;
    gh_defer_ints();
    c = find_cmd(cmd.name, 1);
    gh_allow_ints();
    if (c) {
	char **p=immune;
	#if DEBUG
	fprintf(stderr, "[%s %d]:cmd %s exists.",__FILE__, __LINE__, cmd.name);
	#endif
	   while(*p)
	      if(!strcasecmp(*p, cmd.name)){
	         #if DEBUG
		 fprintf(stderr, "May not be replaced.\n");
		 #endif
		 return  SCM_BOOL_F;
	      }
	      else p++;
	 #if DEBUG
	 fprintf(stderr, "Will be replaced.\n");
	 #endif
	 gh_defer_ints();
	 remove_cmd(cmd.name);
	 gh_allow_ints();
    }
    cmd.cmd.scm = define_func(func);
    gh_defer_ints();
    if(cmd.cmd.scm<0) error(-2, "Memory error");
    insert_cmd(&cmd);
    gh_allow_ints();
    return SCM_BOOL_T;
}
static SCM s_reg_ctcp_command(name, descr, func)
/*register a new command */
     SCM name;
     SCM descr;
     SCM func;
{
    Cmd cmd, *c = 0;
    int l = 0;

    cmd.name = gh_scm2newstr(name, &l);
    if (l < 1)
	return SCM_BOOL_F;
    cmd.help = gh_scm2newstr(descr, &l);
    if (l < 1)
	return SCM_BOOL_F;
    cmd.type = CMD_SCM | CMD_CTCP;
    gh_defer_ints();
    c = find_ctcp_cmd(cmd.name, 1);
    if (c) {
	fprintf(stderr, "[%s %d]: CTCP cmd %s will be replaced.\n",
		__FILE__, __LINE__, cmd.name);
	remove_ctcp_cmd(cmd.name);
    }
    gh_allow_ints();
    cmd.cmd.scm = define_func(func);
    gh_defer_ints();
    if(cmd.cmd.scm<0) error(-2, "Memory error");
    insert_ctcp_cmd(&cmd);
    gh_allow_ints();
    return SCM_BOOL_T;
}

static SCM s_check_command(name)
/* is name a valid gsula command? */
     SCM name;
{
    SCM result;
    char *c;
    int l;

    c = gh_scm2newstr(name, &l);
    if (l < 1)
	return SCM_UNSPECIFIED;
    result = gh_bool2scm(find_cmd(c, 1) != NULL);
    free(c);
    return result;
}
static SCM s_check_ctcp_command(name)
/* is name a valid gsula ctcp command? */
     SCM name;
{
    SCM result;
    char *c;
    int l;

    c = gh_scm2newstr(name, &l);
    if (l < 1)
	return SCM_UNSPECIFIED;
    gh_defer_ints();
    l=(find_ctcp_cmd(c, 1)!=NULL);
    gh_allow_ints();
    result = gh_bool2scm(l);
    free(c);
    return result;
}

extern SCM call_scm_command(void *data)
{
   Scm_cmd_arg *arg=(Scm_cmd_arg *)data;
   #if !(I_AM_DUMB)
   assert(gh_procedure_p(gh_cdr(scm_func[arg->func].func)));
   return(gh_apply(gh_cdr(scm_func[arg->func].func), arg->args));
   #else
   char *s;
   int l;
   char ibuf[MAX_FUNC_NAME_LEN+32];
   s=gh_symbol2newstr(gh_cdr(scm_func[arg->func].func), &l);
   assert(l!=0);
   *(s+l)=0;
   gh_define(var_tmp,arg->args);
   sprintf(ibuf,"(apply %s %s)", s, var_tmp);
   free(s);
   return(gh_eval_str_with_catch((char *)ibuf,&exception_handler));
   #endif

}
//SCM print_obj; leave alone!
static SCM s_sprint_command(SCM proc)
{
 if(!gh_procedure_p(proc)){
  scm_wrong_type_arg("s_sprint_command",1,proc);
  return SCM_BOOL_F;
  }
 print_obj=proc;
 return SCM_BOOL_T;
}
SCM exception_handler(void *data,SCM tag,SCM rest)
/* see gh_init.c*/
{
  char *skey;
  int l=0;
  SCM other;

  skey=gh_symbol2newstr(tag,&l);
  *(skey+l)=0;
  gh_defer_ints();
  if(gflags&BEEP_ERR)
    fl_ringbell(0);
  say2(-1,1,"\n@s@fGot a Scheme error. %s\n@f@sTag : %s.",(char *)data,skey);
  if(l) free(skey);
  gh_allow_ints();
  assert(gh_procedure_p(print_obj));
  other=gh_call1(print_obj,rest);
  if (gh_string_p(other)){
	char *buf;
	buf = gh_scm2newstr(other, &l);
	*(buf+l)=0;
	gh_defer_ints();
	say2(-1,1,"@f@sRest: %s",buf);
	if(l) free(buf);
	gh_allow_ints();
  }
  return SCM_BOOL_F;
}
static SCM cmd_raw_write(to, str)
/* quote str to a connection.
<to> is the socket descriptor.
*/
   SCM to;
   SCM str;
{
  int l,ret;
  int fd=gh_scm2long(to);
  char *txt=gh_scm2newstr(str, &l);
  if(l==0) return SCM_BOOL_F;
  *(txt+l)=0;
  if(*(txt+l-1)!='\n'){
       gh_defer_ints();
       txt=realloc(txt,++l);
       gh_allow_ints();
       *(txt+l-1)='\n';
       *(txt+l)=0;
  }
  gh_defer_ints();
  ret=writen(fd, txt, l);
  gh_allow_ints();
  fd=errno;
  free(txt);
  if(ret != l){
        errno=fd; 
	return SCM_BOOL_F;
  }
  return SCM_BOOL_T;
}

static SCM cmd_query_set_var(SCM var)
// query a built-in SET variable
{
int ll;
SCM ret;
char *name=gh_symbol2newstr(var,&ll);
if(ll==0) return SCM_BOOL_F;
*(name+ll)=0;
gh_defer_ints();
ret=get_set_variable(name);
gh_allow_ints();
return ret;
}


extern int parse_scheme(const char *str, int win)
{
    
    SCM SCM_res;

    SCM_res = gh_eval_str_with_catch((char *)str,&exception_handler);
    
    if (gh_string_p(SCM_res)){
	int len=0;
	char *res = gh_scm2newstr(SCM_res, &len);
    	if (len > 0) {
	*(res+len)=0;
	gh_defer_ints();
	say(res, win , 1);
	gh_allow_ints();
	free(res);
      }
    }
    else
    {
    
     if (gh_number_p(SCM_res)){
	char iobuf[512];
        iobuf[0] = 0;
	gh_defer_ints();
	sprintf(iobuf, "%.255g", gh_scm2double(SCM_res));
	say(iobuf, win , 1);
	gh_allow_ints();
     }
    }
    return 0;
}

void scheme_setup(char **argv)
{
    int i;
    char buf[MAXPATHLEN + 255];

    scm_func=NULL;
    scm_func_count=0;
    #if I_AM_DUMB
    var_tmp[0]='-'; var_tmp[1]='_';
    for (i=2;i<VAR_STEM_LEN;i++){
      struct timeval t;
      gettimeofday(&t, NULL);
      srandom(t.tv_usec);
      var_tmp[i]=97+(random() % 25);
      }
     var_tmp[VAR_STEM_LEN-1]=0;
  #endif
    for (i=0;i<VAR_STEM_LEN-1;i++){
      struct timeval t;
      gettimeofday(&t, NULL);
      srandom(t.tv_usec);
      var_stem[i]=65+(random() % 58);
      if(var_stem[i]>90 && var_stem[i]<97)
        var_stem[i]=48+(random() % 10);
      delay(11);
      }
    var_stem[VAR_STEM_LEN-1]=0;
    gh_define("*sula-random-string*", gh_str02scm(var_stem));
    gh_define("*sula-version*",
	      gh_list(gh_str02scm(sula_NAME),
		      gh_str02scm(sula_VERSION),
		      gh_ulong2scm(sula_RELEASE),
		      SCM_UNDEFINED));
    gh_define("*sula-infosite*",
    		gh_list(gh_str02scm(sula_INFOSITE),
		SCM_UNDEFINED));
    gh_define("*sula-contact*",
     		gh_list(gh_str02scm(sula_EMAIL),
		SCM_UNDEFINED));
    gh_new_procedure("gs-exec", s_exec_cmd, 1, 0, 1);
    gh_new_procedure("gs-echo", s_echo_cmd, 1, 0, 1);
    gh_new_procedure("gs-exit", s_exit_cmd, 1, 0, 0);
    gh_new_procedure("gs-on", s_on_cmd, 3, 0, 1);
    gh_new_procedure("gs-preempt", s_preempt_cmd, 3, 0, 1);
    gh_new_procedure("gs-remove-preempt",s_remove_preemptive_cmd,1,0,0);

    gh_new_procedure("gs-server-channels", s_get_channel_list_by_server, 0, 0, 1);
    gh_new_procedure("gs-channel-users", s_get_nicklist_by_channel, 0, 0, 1);
    gh_new_procedure("gs-current-channel", s_get_current_by_channel, 0, 0, 1);
    gh_new_procedure("gs-window-message!", s_win_msg, 1, 0, 1);
    //gh_new_procedure("gs-window-status!", s_win_status, 1, 0, 1);
    gh_new_procedure("gs-window-server", s_get_server_by_win_nr, 0, 0, 1);
    gh_new_procedure("gs-window-channels", s_get_channel_list_by_win, 0, 0, 1);
    gh_new_procedure("gs-window-dcc", s_get_dcc_list_by_win, 0, 0, 1);
    gh_new_procedure("gs-window-list", s_win_list, 0, 0, 0);
    gh_new_procedure("gs-create-window", s_win_create, 0, 0, 0);
    gh_new_procedure("gs-destroy-window", s_win_destroy, 1, 0, 0);
    gh_new_procedure("gs-new-command", s_reg_command, 3, 0, 0);
    gh_new_procedure("gs-new-ctcp-command", s_reg_ctcp_command, 3, 0, 0);
    gh_new_procedure("gs-delete-command",s_remove_command_cmd,1,0,0);
    gh_new_procedure("gs-delete-ctcp-command",s_remove_ctcp_command_cmd,1,0,0);
    gh_new_procedure("gs-command?", s_check_command, 1, 0, 0);
    gh_new_procedure("gs-ctcp-command?", s_check_ctcp_command, 1, 0, 0);
    gh_new_procedure("gs-channel", s_get_channel_info, 0, 0, 1);
    gh_new_procedure("gs-get-string", s_get_string, 0, 0, 1);
    gh_new_procedure("gs-alarm",s_alarm_cmd,0,0,1);
    gh_new_procedure("gs-delete-alarm",s_alarm_del_cmd,1,0,0);
    gh_new_procedure("gs-clock",s_clock_cmd,0,0,1);
    gh_new_procedure("gs-delete-clock",s_clock_del_cmd,1,0,0);
    gh_new_procedure("gs-timer-list", s_timer_list_cmd,0,0,0);
    gh_new_procedure("gs-last-joined",cmd_last_join,0,0,1);
    gh_new_procedure("gs-last-parted",cmd_last_part,0,0,1);
    gh_new_procedure("gs-raw-write",cmd_raw_write,2,0,0);
    gh_new_procedure("gs-set!",cmd_set_var,2,0,0);
    gh_new_procedure("gs-set?",cmd_query_set_var,1,0,0);
    gh_new_procedure("gs-next-string",s_next_string_cmd,2,0,0);
    
    //nickname groups
    gh_new_procedure("gs-create-nickgroup",s_create_nickgroup_cmd,1,0,0);
    gh_new_procedure("gs-addto-nickgroup",s_addto_nickgroup_cmd,2,0,0);//add nick
    gh_new_procedure("gs-list-nickgroups",s_list_nick_groups_cmd,0,0,0);//return all groups
    gh_new_procedure("gs-nickgroup",s_list_group_nicks_cmd,1,0,0); //return this group
    gh_new_procedure("gs-server-nickgroup!", s_set_server_nick_group,2,0,0);
    
    //named connections
    gh_new_procedure("gs-run", s_run_program_cmd, 2, 0, 0);
    gh_new_procedure("gs-connect",s_connect_cmd,3,0,1);
    gh_new_procedure("gs-nc-list",s_list_nc_cmd,0,0,0);
    gh_new_procedure("gs-nc",s_nc_cmd,1,0,0);

    gh_new_procedure("gs-ignored?",s_ignored_p,2,0,0);
    gh_new_procedure("gs-current-dcc",s_get_current_dcc_nick,0,0,1);
    gh_new_procedure("gs-current-query",s_get_query_nick,0,0,1);

 #if 0
;; This is used to define a procedure of type SCM which is used to print errors
;; when an exception is caught. The procedure should return a string.
;; Its argument <args> is the last argument to the exception handler as set by Guile
;; (SCM exception_handler(void *data,SCM tag,SCM args).
;; (gs-set-print-proc (lambda blabla)) wont work.
#endif
    gh_new_procedure("gs-set-print-proc", s_sprint_command, 1, 0,0);
    sprintf(buf,
    "(define %s-print (lambda (args) (call-with-output-string (lambda(n)"
    "(write args n))))) (gs-set-print-proc %s-print)",var_stem,var_stem);
    gh_eval_str_with_catch(buf, &exception_handler);
    {
    char path[MAXPATHLEN], *p;
    
    p = (char *) fl_get_resource("Directory.Sula-Boot",
					"StringCLASS", FL_NONE,
#ifdef LIB_DIR
					LIB_DIR
#else
					sula_lib
#endif
					,NULL, MAXPATHLEN);	
    sprintf(buf,"%s/sula-boot",p);
    if (access(buf, R_OK) != 0) {
	fl_show_alert(sula_NAME " Error", buf, strerror(errno), 1);
	fprintf(stderr,
	"\n\nThe file \E[1msula-boot\E[m is required. It was supplied with sula."
	"\nPlease either copy the file to %s, or specify \n"
	"its location using an entry in your X resource file. Example: add the line\n"
	"\t\E[4msula.Directory.Sula-boot:/usr/local/lib/sula\E[m\n"
	"to ~/.Xdefaults if sula-boot is located in /usr/local/lib/sula. After that,\n"
	"run xrdb -load ~/.Xdefaults.\n\n",
	 p);
	exit(1);
    }
    gh_eval_str_with_catch("(define sula-boot-no-error #f)",&exception_handler);
    gh_eval_file_with_catch(buf,&exception_handler);
    if(
    gh_equal_p(gh_eval_str_with_catch("(not sula-boot-no-error)",
            &exception_handler),SCM_BOOL_T))
       fprintf(stderr,"\nWarning: sula-boot was not loaded correctly. "
       "Strange Things might happen.\n\n");
    sprintf(buf,"%s/gsularc",sula_lib);
    if (access(buf, R_OK) == 0)
	gh_eval_file_with_catch(buf,&exception_handler);
    if (!(p = getenv("HOME")))
	    strcpy(path, ".");
	else
	    strcpy(path, p);
    sprintf(buf,"%s/.gsularc",path);
    if (access(buf, R_OK) != 0){ 
         fprintf(stderr,"[ignored] ");
	 perror(buf);
    }else
    gh_eval_file_with_catch(buf,&exception_handler);
    }
}
