#include <forms.h>
#include <errno.h>
#include <sys/time.h>
/*#include <stdarg.h> */
/*#include <pwd.h> */
#include <ctype.h>
#include <assert.h>
#include <netdb.h>
#include <arpa/inet.h>
#include "config.h"
#include "nicks.h"
#include "aux.h"
#include "sula.h"
#include "servers_gui.h"
#include "chanwin.h"
#include "global.h"
#include "lag.h"
#include "cmd.h"
#include "hooks.h"
FD_main_form *main_form;
extern FD_serverlist *server_tool;

static int del_server(pServer * pp, Server * s, int flag);
static void print_server_tree(Server * p);
int load_serverlist(const char *filename)
{
    Server server;
    FILE *fp;
    int i;
    static int get_server(FILE *, Server *);

    if (!(fp = fopen(filename, "r")))
	return errno;
    while ((i = get_server(fp, &server)) != EOF)
	if (i == 0) {
	    insert_server(&servTree, &server, BY_NAME);
	}
#if DEBUG
	else if (i == 2)
	    fputs("invalid server entry", stderr);
#endif
    fclose(fp);
    return 0;
}
int save_servers(char *path)
{
    static void save_server_list(FILE * fp, Server * p);
    char *buf;
    FILE *fp;

    if (path)
	goto do_it;
    if (!(path = getenv("HOME")))
	path = ".";
    buf = (char *) malloc(sizeof(char) * (strlen(path) + strlen(P_HOME P_SERVERS) + 8));

    sprintf(buf, "%s/%s/%s", path, P_HOME, P_SERVERS);
    path = strdup(buf);
    if (access(path, F_OK) > -1) {
	if (access(path, W_OK) < 0) {
	    free(buf);
	    return -1;
	}
	sprintf(buf, "%s.bak", path);
	rename(path, buf);
    }
    free(buf);
  do_it:
    if ((fp = fopen(path, "w")) == NULL)
	return -1;
    else{
    time_t t= time(NULL);
    fprintf(fp, "%c" sula_NAME sula_VERSION "\n%cServer list created %s\n",
     COMMENT, COMMENT, ctime(&t));
    save_server_list(fp, servTree);
    fflush(fp);
    }
    fclose(fp);
    return 0;
}
static void save_server_list(FILE * fp, Server * p)
{
    char buf[512];

    if (p) {
	save_server_list(fp, p->left);
	sprintf(buf, "%s:%u:%s:%s:%s:%s\n", p->servname, p->port, p->password ? p->password : "",
		p->nick, p->email, p->ircname);
	fputs(buf, fp);
	save_server_list(fp, p->right);
    }
}

int remove_server_from_tree(char *servname, int port)
{
    Server *server;

    server = search_server_tree_byname(servname, port);
    if (!server)
	return -1;
    else {
	int found;

	if (search_winlist_byservname(server->servname, server->port, &found))
	    return 0;
	close_server_connection(&server->fd);
	del_server(&servTree, server, BY_NAME);
    }
    return 1;
}

static void left_rotate(pServer * pp)
{
    pServer p = *pp, r;

    *pp = r = p->right;
    p->right = r->left;
    r->left = p;
    p->bf--;
    if (r->bf > 0)
	p->bf -= r->bf;
    r->bf--;
    if (p->bf < 0)
	r->bf += p->bf;
}

static void right_rotate(pServer * pp)
{
    pServer p = *pp, l;

    *pp = l = p->left;
    p->left = l->right;
    l->right = p;
    p->bf++;
    if (l->bf < 0)
	p->bf -= l->bf;
    l->bf++;
    if (p->bf > 0)
	l->bf += p->bf;
}

int insert_server(pServer * pp, Server * s, int flag)
{
    int dH = 0;
    pServer p = *pp;
    int tt = 0;

    assert(s && s->servname && (flag == BY_FD || flag == BY_NAME));
    assert(!(flag == BY_FD && s->fd < 0));
    if (p == NULL) {

	*pp = p = (pServer) malloc(sizeof(Server));
	if (!p)
	    error(1, "malloc error");
	if (flag == BY_FD) {
	    *p = *s;
	    p->port = s->port;
	    p->fd = s->fd;
	    p->servname = strdup(s->servname);
	    p->left = p->right = NULL;
	    p->bf = 0;
	    return 1;
	}
	p->servname = strdup(s->servname);
	p->name = s->name ? strdup(s->name) : NULL;
	p->nick = s->nick ? strdup(s->nick) : "sula";	/* * must be a bug somewhere                                                             */
	p->iNick = s->iNick;
	if (p->iNick < 0) {
	    Nickcount i = get_list_bylabel("default");

	    if (i < 0) {
		i = new_nicklist("default");
		if (i < 0)
		    error(-2, "new_nicklist error");
	    }
	    p->iNick = i;
	    aNickList[p->iNick].current =
	       add_nick_to_list(p->iNick, p->nick);
	    if (!aNickList[p->iNick].current)
		error(-2, "aNickList error");
	}

	p->email = s->email ? strdup(s->email) : "sula@pope.va";	/*wont  happen */
	p->ircname = s->ircname ? strdup(s->ircname) : sula_NAME sula_VERSION " Tester";
	p->mode = NULL;
	p->lag = NULL;
	p->flag = 0;
	p->connecting = 0;
	p->t.tv_sec = p->t.tv_usec = 0;
	p->port = s->port;
	p->state = s->state;
	p->password = strdup2(s->password);
	p->fd = s->fd;
	p->lastInvite = NULL;
	p->read = 0;
	p->wrote = 0;
	p->connect_time=0;
	p->bf = 0;

	p->left = p->right = NULL;
	dH = 1;
    }
    else if (
		(flag == BY_NAME &&
		 ((tt = strcasecmp(s->servname, p->servname)) > 0 ||
		  (tt == 0 && s->port > p->port))
		) ||
		(flag == BY_FD && s->fd > p->fd)
       ) {
	if (insert_server(&p->right, s, flag)) {
	    p->bf++;
	    if (p->bf == 1)
		dH = 1;
	    else if (p->bf == 2) {
		if (p->right->bf == -1)
		    right_rotate(&p->right);
		left_rotate(pp);
	    }
	}
    }
    else if (
	    (flag == BY_NAME && (tt < 0 || (tt == 0 && s->port < p->port))) ||
		(flag == BY_FD && s->fd < p->fd)
       ) {
	if (insert_server(&p->left, s, flag)) {
	    p->bf--;
	    if (p->bf == -1)
		dH = 1;
	    else if (p->bf == -2) {
		if (p->left->bf == 1)
		    left_rotate(&p->left);
		right_rotate(pp);
	    }
	}
    }
    else if (flag == BY_FD && s->fd == p->fd) {
	/*two servers have same fd */
	error(0, "[%s %d] Error", __FILE__, __LINE__);
	raise(SIGSEGV);
    }
    return dH;
}

static int del_server(pServer * pp, Server * s, int flag)
{
    pServer p = *pp, *q;
    int dH = 0;
    int tt = 0;

    if (!p)
	return 0;
    if (
	   (flag == BY_FD && s->fd < p->fd) ||
	   (flag == BY_NAME && (
			    (tt = strcasecmp(s->servname, p->servname)) < 0 ||
				   (tt == 0 && s->port < p->port)
	    )
	   )
       ) {
	if (del_server(&p->left, s, flag)) {
	    p->bf++;
	    if (p->bf == 0)
		dH = 1;
	    else if (p->bf == 2) {
		if (p->right->bf == -1)
		    right_rotate(&p->right);
		left_rotate(pp);
		if (p->bf == 0)
		    dH = 1;
	    }
	}
    }
    else if ((flag == BY_FD && s->fd > p->fd) ||
	     (flag == BY_NAME && (tt > 0 || (tt == 0 && s->port > p->port)))
       ) {
	if (del_server(&p->right, s, flag)) {
	    p->bf--;
	    if (p->bf == 0)
		dH = 1;
	    else if (p->bf == -2) {
		if (p->left->bf == 1)
		    left_rotate(&p->left);
		right_rotate(pp);
		if (p->bf == 0)
		    dH = 1;
	    }
	}
    }
    else {			/* * found  */
	if (p->right == NULL) {
	    *pp = p->left;
	    if (flag == BY_FD) {
		free(p);
	    }
	    else {		/* * do_free(p); */
#ifdef this_will_be_fixed
		free(p->mode);
		free(p->servname);
		free(p->name);
		etc
#endif
		   free(p);
	    }
	    return 1;
	}
	else if (p->left == NULL) {
	    *pp = p->right;
	    if (flag == BY_FD) {
		free(p);
	    }
	    else {		/*       * do_free(p); */
#ifdef this_will_be_fixed
		free(p->servname);
		free(p->name);
		free(p->nick);
		etc
#endif
		   free(p);
	    }
	    return 1;
	}
	else {
	    q = &p->left;
	    while ((*q)->right != NULL)
		q = &(*q)->right;
	    if (flag == BY_FD) {
		p->fd = (*q)->fd;
		p->servname = (*q)->servname;
		p->name = (*q)->name;
		p->port = (*q)->port;

	    }
	    else {
		p->port = (*q)->port;
		p->servname = strdup2((*q)->servname);
		p->name = strdup2((*q)->name);
		p->fd = (*q)->fd;
		p->nick = strdup2((*q)->nick);
		p->ircname = strdup2((*q)->ircname);
		p->mode = strdup2((*q)->mode);
		p->flag = (*q)->flag;
		p->connecting = (*q)->connecting;
		p->t.tv_sec = (*q)->t.tv_sec;
		p->t.tv_usec = (*q)->t.tv_usec;
		p->lag = (Lag *) malloc(sizeof(Lag));
		p->lag = ((*q)->lag);
		p->password = strdup2((*q)->password);
		p->email = strdup2((*q)->email);
		p->state = (*q)->state;
		p->read = (*q)->read;
		p->wrote = (*q)->wrote;
		p->connect_time = (*q)->connect_time;
		p->lastInvite = strdup2((*q)->lastInvite);
		p->iNick = (*q)->iNick;
	    }
	    if (flag == BY_FD) {
		(*q)->fd = s->fd;
		(*q)->servname = s->servname;
		(*q)->port = s->port;
	    }
	    else {

		(*q)->fd = s->fd;
		(*q)->servname = strdup(s->servname);
		(*q)->name = strdup2(s->name);
		(*q)->nick = strdup2(s->nick);
		(*q)->ircname = strdup2(s->ircname);
		(*q)->email = strdup2(s->email);
		(*q)->flag = s->flag;
		(*q)->connecting = s->connecting;
		(*q)->t.tv_sec = s->t.tv_sec;
		(*q)->t.tv_usec = s->t.tv_usec;
		(*q)->lag = (Lag *) malloc(sizeof(Lag));
		(*q)->lag = s->lag;
		(*q)->mode = strdup2(s->mode);
		(*q)->port = s->port;
		(*q)->password = strdup2(s->password);
		(*q)->state = s->state;
		(*q)->read = s->read;
		(*q)->wrote = s->wrote;
		(*q)->connect_time = s->connect_time;
		(*q)->lastInvite = strdup2(s->lastInvite);
		(*q)->iNick = s->iNick;
	    }
	    if (del_server(&p->left, s, flag)) {
		p->bf++;
		if (p->bf == 0)
		    dH = 1;
		else if (p->bf == 2) {
		    if (p->right->bf == -1)
			right_rotate(&p->right);
		    left_rotate(pp);
		    dH = 1;
		}
	    }
	}
    }
    return dH;
}

static void print_tree(Server * p, int pos)
{
    if (p) {
	print_tree(p->right, pos + 8);
	printf("%*.8s(%d) %d\n", pos, p->servname, p->port, p->bf);
	print_tree(p->left, pos + 8);
    }
}

static void print_fdtree(Server * p, int pos)
{
    if (p) {
	print_fdtree(p->right, pos + 8);
	printf("%*d %6s %d \n", pos, p->fd, p->servname, p->bf);
	print_fdtree(p->left, pos + 8);
    }
}

void list_servers(void)
{
    print_server_tree(servTree);
    print_tree(servTree, 8);
    print_fdtree(servFdTree, 3);
}

void print_server_tree(Server * p)
{
    int i = 1;

    if (p) {
	print_server_tree(p->left);
	fprintf(stderr, "%s\t%d. %s %d %s %s (%s)",
		p->fd > 0 ? "*" : "", i++, p->servname,
		p->port, p->nick, p->email, p->ircname);
	print_server_tree(p->right);
    }
}
static Server *
    search_server_tree_byfd(int fd)
{
    Server *p = servFdTree;

    while (p) {
	if (fd == p->fd)
	    return p;
	if (fd < p->fd)
	    p = p->left;
	else
	    p = p->right;
    }
    return NULL;
}
Server *
    search_server_tree_byname(const char *name, int port)
{
    Server *p = servTree;
    int c;

    while (p) {
	c = strcasecmp(name, p->servname);
	if (c == 0) {
	    if (p->port == port)
		return p;
	    else if (port > p->port)
		p = p->right;
	    else
		p = p->left;
	}
	else if (c < 0)
	    p = p->left;
	else
	    p = p->right;
    }
    return NULL;
}
static void RemoveChannels(Server *);
static void RemoveCommands(Server *);
static void close_server(Server * server)
{
    del_server(&servFdTree, server, BY_FD);
    UpdateWindows(server->fd, 0);
    ClearLag(server);
    server->fd = -1;
    server->flag&=~SET_AWAY;
    if (server->connecting > 0)
	server->connecting = 0;	/* dont reset if <0 : ext proc. */
    RemoveCommands(server);
    RemoveChannels(server);
    if (server_tool)
	fl_trigger_object(server_tool->update);
}

static void ircserver_cb(int fd, void *data)
{
#define SERV_RD_MAX	1024
    char line[SERV_RD_MAX + 1];
    int n;

#ifndef SA_RESTART
  retry:
#endif
    if ((n = read(fd, line, SERV_RD_MAX)) < 0) {
	if (errno != EAGAIN)
#ifndef SA_RESTART
	    if (errno == EINTR)
		return;		/* actually should be: goto retry; */
	    else
#endif
	    {
		Winstruct *win;

		fl_remove_io_callback(fd, FL_READ, ircserver_cb);
		close(fd);
		win = search_winlistbyfd(fd, &n);
		assert(win != NULL);
		if (!check_msg_hook(SERVER_LOST, win - winstruct, "%s,%hu %s %s",
		      win->server->servname, win->server->port,win->server->nick,
		      win->server->name? win->server->name:win->server->servname))
		    say2(-1, 1, PROMPT "Read error from port %hu of server %s.Closing...",
			 win->server->port, win->server->name);
		close_server(win->server);
	    }
	return;
    }
    if (n == 0) {
	Winstruct *win;

	fl_remove_io_callback(fd, FL_READ, ircserver_cb);
	close(fd);
	win = search_winlistbyfd(fd, &n);
	assert(win);
	if(!check_msg_hook(SERVER_LOST, win - winstruct, "%s,%hu %s %s",
		       win->server->servname,
		       win->server->port, win->server->nick,
		        win->server->name? win->server->name:win->server->servname))
	    say2(-1, 1, PROMPT "Connection closed from port %hu of server %s",
		 win->server->port, win->server->name?win->server->name:
		 win->server->servname);
	close_server(win->server);
	return;
    }
    line[n] = '\0';
    prepare_server_output(fd, line);

}
/********************************************************/
/*                                                      */
/*      C o n n e c t   t o   a   s e r v e r           */
/*                                                      */
/********************************************************/

extern void connecting(char *server, u_short port, char done)
{
/* to call before and after successfull connection to server */
    WinList *p;

    build_winlistbyservname(server, port);
    p = servWin;
    while (p) {
	if (!done) {
	    fl_deactivate_object(p->win->chanwin->close_con);
	    fl_set_object_color(p->win->chanwin->close_con,
				FL_INACTIVE_COL, FL_MCOL);
	}
	else {
	    fl_activate_object(p->win->chanwin->close_con);
	    fl_set_object_color(p->win->chanwin->close_con,
				FL_COL1, FL_MCOL);
	}
	p = p->next;
    }
    if (done && server_tool)
	fl_trigger_object(server_tool->update);

}

typedef struct {
    char *name;			/*server name */
    u_short port;
    int win;
} Hp_struct;
extern int RemoveFromWinChanList(Winstruct *, const char *);
static void gethp_cb(int fd, void *data)
/* accept server IP number */
{
#define MAXRD	255
    char line[MAXRD + 1], *buf = 0;
    int n;
    Hp_struct *hp = data;
    Winstruct *win = (hp->win > -1 && winstruct[hp->win].chanwin) ?
    &winstruct[hp->win] : NULL;
    long old = 0;
    int no_of_wins = 0, old_fd = -1;
    Server *s;

    n = read(fd, line, MAXRD);
    fl_remove_io_callback(fd, FL_READ, gethp_cb);
    close(fd);
    if (server_tool)
	fl_trigger_object(server_tool->update);
    if (win) {
	if (win->chanwin->close_con->u_vdata)
	    win->chanwin->close_con->u_vdata = NULL;
	else {
	    /* another serevr is has connected using our win */
#if DEBUG
	    fprintf(stderr, "too late (%p)..?\n", win);
#endif
	    return;
	}
	old = (long) win->server;
	if (win->server) {
	    search_winlistbyfd(win->server->fd, &no_of_wins);
	    old_fd = win->server->fd;
	}
    }
    s = search_server_tree_byname(hp->name, hp->port);
    if (s == NULL) {		/* server's beeen removed in the meantime!!!! */
	fl_ringbell(25);
#if DEBUG
	fprintf(stderr, "server %s (%hu) is gone..?\n", hp->name, hp->port);
#endif
	free(hp->name);
	return;
    }
    free(hp->name);
    assert(s->connecting == -1);
    s->connecting = 0;
    buf = malloc(sizeof(char) * (strlen(s->servname) + 255));

    if (n < 1) {
	sprintf(buf, "Process couldnt get hostname for %s so stop.",
		s->servname);
	    goto out;
    }
    else {
	line[n] = '\0';
	sprintf(buf, "Got hostname for %s:%s."
		" Trying to connect...", s->servname, line);
    }
    if (win) {
	fit_object_label(win->chanwin->WinStatusLine, buf);
	XFlush(fl_display);
    }
    if (server_tool)
	fit_object_label(server_tool->status, buf);
#if DEBUG
    error(0, "%s", buf);
#endif
    free(buf);
    if (inet_addr(line) == INADDR_NONE) {	/* external prog buggy or what? */
	error(0, "Bad IP address: %s. Stop.", line);
	line[0] = 0;
    }
    if (connect_to_server(s, win, line[0] ? line : NULL, 1)) {
	buf = malloc(sizeof(char) * (strlen(s->servname) + 80));

	sprintf(buf, "Connection to port %d of %s failed",
		s->port, s->servname);
	if (win)
	    win->server = (Server *) old;
      out:
	if (win)
	    fit_object_label(win->chanwin->WinStatusLine, buf);
#if DEBUG
	error(0, "%s", buf);
#endif
	if (server_tool)
	    fl_set_object_label(server_tool->status, buf);
	free(buf);
	return;
    }
    else if (win && old) {
	if (no_of_wins > 1) {
	    if (old_fd > -1) {
		Channel *current = win->chanStart->right;
		while (current != win->chanStart) {
		    sprintf(line, "NOTICE %s :(%s changing server for %p)\n",
			    current->name, sula_NAME sula_VERSION, win);
		    n = strlen(line);
		    if (writen(old_fd, line, n) != n)
			goto ende;
		    sprintf(line, "PART %s\n", current->name);
		    n = strlen(line);
		    if (writen(old_fd, line, n) != n)
			goto ende;
		    current = current->right;
		    continue;
		  ende:
		    close_server_connection(&old_fd);
		    break;
		}
	    }
	}
	if ((Server *) old != win->server && no_of_wins < 2) {
	    new_close_connection(&((Server *) old)->fd,
			      sula_NAME sula_VERSION ": Changing servers", 0);
	    ((Server *) old)->state = -1;
	}
    }				/* else if (win && old) */
}

static jmp_buf env;
static void sig_alrm(int signo)
{
    longjmp(env, 1);
}

static int gethostname_for_connection(Server * s, Winstruct * win)
{
/* determine server IP number */
    int fd[2], pid;

    if (win)
	fit_object_label(win->chanwin->WinStatusLine,
			 "Resolving server name. Bitte warten Sie");

#if DEBUG
    if (gflags & VERBOSE_CLIENT)
	error(0, "Starting process for remote hostname %s", s->servname);
#endif
    if (socketpair(AF_UNIX, SOCK_STREAM, 0, fd) < 0) {
	error(-1, "socketpair error");
	return -1;
    }
    TELL_WAIT();		/* to synchronize child/parent */
    if ((pid = fork()) < 0) {
	error(-1, "fork error");
	return -1;
    }
    else if (pid > 0) {
	Hp_struct *hp = malloc(sizeof(Hp_struct));

#if DEBUG
	if (gflags & VERBOSE_CLIENT)
	    if (sflags & EXTERNALGETHOST)
		error(0,
		   "[%s] pid %d. Timeout set to %d secs (/set tmout_connect)."
		      " Bitte etwas Geduld...",
		      gethostbynameprogram, pid, connect_timeout);
#endif
	hp->name = strdup(s->servname);
	hp->port = s->port;
	hp->win = -1;
	if (win) {
	    register i;

	    for (i = 0; i < win_count; i++)
		if (&winstruct[i] == win) {
		    hp->win = i;
		    win->chanwin->close_con->u_vdata = (void *) hp;	/*
									   anything that'll let us know that we are connecting */
		    break;
		}
	}
	close(fd[1]);
	fl_add_io_callback(fd[0], FL_READ, gethp_cb, (void *) hp);
	TELL_CHILD(pid);
	return 0;
    }
    else {
	int res = -1;

	WAIT_PARENT();		/* wait for parent to set up callback */
	close(fd[0]);
	if (fd[1] != STDOUT_FILENO)
	    if (dup2(fd[1], STDOUT_FILENO) != STDOUT_FILENO)
		exit(-1);
	if (sflags & EXTERNALGETHOST) {
	    char buf2[12];
	    char *prog = strrchr(gethostbynameprogram, '/');

	    if (!prog || !(*++prog))
		prog = gethostbynameprogram;
	    sprintf(buf2, "%u", connect_timeout);
	    res = execlp(gethostbynameprogram, prog, "-t",
			 buf2, s->servname, NULL);
	}
#if 1
	if (res == -1) {
	    // now try again
	    struct hostent *hp;

	    if (sflags & EXTERNALGETHOST) {
		perror(gethostbynameprogram);
		fprintf(stderr, "Trying again with gethostbyname.\n");
	    }
	    if (connect_timeout) {
		if (signal(SIGALRM, sig_alrm) == SIG_ERR)
		    perror("signal"),
		       exit(-1);
		if (setjmp(env) != 0)
		    fprintf(stderr, PROMPT "timeout resolving hostname\n"),
		       exit(-2);
		alarm(connect_timeout);
	    }
	    hp = gethostbyname(s->servname);
	    if (connect_timeout)
		alarm(0);
	    if (hp == NULL)
		herror("gethostbyname"), exit(-1);
	    printf("%s", inet_ntoa(*((struct in_addr *) *hp->h_addr_list)));
	}
	exit(0);
#endif
    }
    return 0;
}

static void rm_cmdoncon(Server * server)
{
/* remove all commands to be done upon connecting to this server */
    CmdOnCon *p = cmdonconStart;

    while (p != cmdonconEnd)
	if ((!p->server && (p->port == server->port || p->port < 1)) ||
	    (!strcasecmp(p->server, server->servname) &&
	     (p->port == server->port || p->port < 1))
	   ) {
	    CmdOnCon *q = p->next;

	    free(p->cmd);
	    free(p->server);
	    if (q == cmdonconEnd)
		cmdonconEnd = p;
	    else
		*p = *q;
	    free(q);
	}
	else
	    p = p->next;
}

extern int connect_to_server(Server * s, Winstruct * win,
			     const char *ip_addr, int have_tried)
{
    /*
     * have to sort all these out someday!
     * -1: already conected
     * 0:  either success (2nd time we get in here), or calling ext. process
     * errno: some error..
     * 
     * create new win if win==NULL
     */
    char *win_title = 0;
    char line[512];
    Server *server = NULL;
    int fd;
    int found = 0, new_win;
    int n, err_no;
    struct linger l;

    assert(s && s->servname);	/* uh? cant fail..? */
    if (ip_addr || have_tried)
	goto do_it;		/* either we've been here before or we 
				   already have an IP address */
    server = search_server_tree_byname(s->servname, s->port);
    if (server) {
	if (server->connecting) {
	    say2(win ? win - winstruct : -1, 1,
		 "already trying to resolve server name on this window:"
		 "%s (%hu). Wait or use another window.\n", server->servname, server->port);
	    return -2;
	}
	else {
	    found = 1;
	    if (server->fd > -1) {
		fl_ringbell(0);
		say2(win ? win - winstruct : -1, 1,
		     "already connected to %s (%hu)\n",
		     server->servname, server->port);
		return -1;
	    }
	}
    }
    else {
	insert_server(&servTree, s, BY_NAME);
	server = search_server_tree_byname(s->servname, s->port);
	assert(server);
    }
    if (inet_addr(server->servname) != INADDR_NONE)
	goto do_it;
    if (1) {
	if (win && win->server) {
	    Winstruct *w;

	    w = search_winlist_byservname(server->servname, server->port, &n);
	    if (w && (w->server->port != win->server->port ||
		    strcasecmp(w->server->servname, win->server->servname))) {
		sprintf(line, "Window exists for %s [%hu]. Switching to it...",
			w->server->servname, w->server->port);
		fit_object_label(win->chanwin->WinStatusLine, line);
		XFlush(fl_display);
		win = w;
	    }
	}
	server->connecting = -1;
	if (gethostname_for_connection(server, win) == 0) {
	    if (server_tool)
		fl_trigger_object(server_tool->update);
	    return 0;		/* return value now meaningless to caller.
				   sit back and wait. we shall be back with the IP numbers */
	}
	else {
	    server->connecting = 0;

	    error(0, "Coulnt start process. Connecting to %s (%hu)"
		  " but may block", server->servname, server->port);
	}
    }
    else {
	error(0, "Attempting to connect to %s (%hu) but may block.",
	     server->servname, server->port);
    }
  do_it:
    if (have_tried) {
	found = 1;
	server = search_server_tree_byname(s->servname, s->port);
	assert(server);		/* something terribly wrong if server not found */
    }
#if DEBUG
    error(0, "Connecting to %s. Bonne chance.",
	  ip_addr ? ip_addr : server->servname);
#endif
    if ((fd = tcp_open(ip_addr ? (char *) ip_addr : server->servname,
		       0, NULL, server->port, connect_timeout)) < 0) {
	say2(-1, 1, "@i" PROMPT "%s (%hu) Connection failed. Diagnostics:",
	     server->servname, server->port);
	say(tcp_err_msg, -1, 1);
#if DEBUG
	fprintf(stderr, "%s\n", tcp_err_msg);
#endif
	if (server_tool)
	    fit_object_label(server_tool->status,
			     "Connection failed");
	rm_cmdoncon(server);
	return -1;
    }
    server->fd = fd;
    insert_server(&servFdTree, server, BY_FD);
    assert(search_server_tree_byfd(server->fd));
#if DEBUG
    error(0, "Connected to port %hu of %s (fd=%d)", server->port,
	  server->servname, server->fd);
#endif
    if (fcntl(server->fd, F_SETFL, O_NONBLOCK) == -1) {
	err_no = errno;
	error(-1, "F_SETFL fcntl error");
	close(server->fd);
	rm_cmdoncon(server);
	return err_no;
    }
    l.l_onoff = 0;
    l.l_linger = 0;
    if (setsockopt(server->fd, SOL_SOCKET, SO_LINGER, (void *) &l, sizeof(l)) < 0)
	error(-1, "(ignored) SO_LINGER- setsockopt error");
    new_win = win ? 0 : 1;
    if (!found || server->state < 0) {
	if (win == NULL) {
	    win = new_chan_win();
	    set_object_attribs(win);
	    win->server = server;
	}
	else
	    win->server = server;
	win->server->read = 0;
    }
    else {
	/* check whether server is actually in the server list */
	win = search_winlist_byservname(server->servname, server->port, &n);
	assert(win && n);
    }
    if (win->server->state < 0) {
	win_title = (char *) malloc(sizeof(char) *
	       (strlen(win->server->servname) + strlen(sula_NAME sula_VERSION)
		+ 24));

	sprintf(win_title, "[%d] " sula_NAME sula_VERSION ": %s %d",
	     (win - winstruct) + 1, win->server->servname, win->server->port);
	if (new_win)
	    show_chanwin(win, &win_title, 1, 0);
	else {
	    fl_set_form_title(win->chanwin->chanwin, win_title);
	    free(win_title);
	}
	win->server->state = 1;
    }
    fl_add_io_callback(win->server->fd, FL_READ, ircserver_cb, (void *) 0);

    sprintf(line, "NICK %s\n", win->server->nick);
    n = strlen(line);
    if (writen(win->server->fd, line, n) != n)
	goto do_error;
    if (win->server->password) {
	sprintf(line, "PASS %s\n", win->server->password);
	n = strlen(line);
	if (writen(win->server->fd, line, n) != n)
	    goto do_error;
    }
    if (server->email) {
	char *name = strxstr(win->server->email, "@");
	char *hhost = strsstr(win->server->email, "@");

	if (!name)
	    name = Name;
	if (!hhost)
	    hhost = host;
	sprintf(line, "USER %s %s %s :%s\n", name, hhost,
		win->server->servname, win->server->ircname ?
		server->ircname : sula_NAME sula_VERSION);
    }
    else
	sprintf(line, "USER %s %s %s :%s\n", Name, host, server->servname,
		server->ircname ? server->ircname : sula_NAME sula_VERSION);
    n = strlen(line);
    if ((n = writen(win->server->fd, line, n)) != n)
	goto do_error;
    fit_object_label(win->chanwin->WinStatusLine,
		     "Server contacted. Wait for confirmation..");
    if (server_tool) {
	fl_set_object_label(server_tool->status, "Server contacted");
	fl_trigger_object(server_tool->update);
    }
    connecting(win->server->servname, win->server->port, 0);
    win->server->connecting = 1;
    win->server->connect_time=time(NULL);
    /*if(!(win->server->flag & GOT_MOTD))
       win->server->flag |= GOT_MOTD; *//* confusing */
    gettimeofday(&win->server->t, NULL);
#ifndef NOLAG
    send_lag_junk(win->server, 1, 0);
#endif
    if (new_win)
	show_chanwin(win, &win_title, 0, 1);
    return 0;

  do_error:
    {
	err_no = errno;
	perror("write error");
	close_server_connection(&win->server->fd);
	return (errno = err_no);
    }

}
static void RemoveChannels(Server * server);
void new_close_connection(int *fd, char *reason, int update_win)
{
    /*
     * undate_win: update corresponding channelr windows after closing connection
     */
    Server *s, *server;

    if (*fd < 1)
	return;
    fl_remove_io_callback(*fd, FL_READ, ircserver_cb);
    s = search_server_tree_byfd(*fd);
    assert(s && s->fd == *fd);
    server = search_server_tree_byname(s->servname, s->port);
    assert(!strcasecmp(server->servname, s->servname) && server->port == s->port);
    if (1) {
	char *buf = (char *) malloc(sizeof(char) * (
			 (reason ? strlen(reason) : strlen(quit_text)) + 10));

	if (buf) {
	    sprintf(buf, "QUIT :%s\n", reason ? reason : quit_text);
	    writen(server->fd, buf, strlen(buf));
	    free(buf);
	}
    }
    close(server->fd);
    del_server(&servFdTree, server, BY_FD);
    if (update_win)
	UpdateWindows(*fd, 0);
    ClearLag(server);
    *fd = server->fd = -1;
    if (server->connecting > 0)
	server->connecting = 0;
    if (server_tool)
	fl_trigger_object(server_tool->update);
    RemoveCommands(server);
    RemoveChannels(server);
    return;
}
void close_server_connection(int *fd)
{
    static void RemoveSender(Server *);
    Server *server, *s2;

    if (*fd < 0)
	return;
    fl_remove_io_callback(*fd, FL_READ, ircserver_cb);
    /*
     * do a reverse server  check..make sure we actually have a server connected 
     */
    s2 = search_server_tree_byfd(*fd);
    assert(s2 && s2->fd == *fd);
    server = search_server_tree_byname(s2->servname, s2->port);
    assert(!strcasecmp(server->servname, s2->servname) && server->port == s2->port);
    if (gflags & VERBOSE_CLIENT)
	say2(-1, 1, PROMPT "Sending %s (%d) the QUIT cmd..",
	     server->name ? server->name : server->servname, server->port);
    if (1) {
	char *buf = (char *) malloc(sizeof(char) * (strlen(quit_text) + 10));

	if (buf) {
	    sprintf(buf, "QUIT :%s\n", quit_text);
	    writen(server->fd, buf, strlen(buf));
	    free(buf);
	}
    }
    close(server->fd);
    UpdateWindows(*fd, 0);
    ClearLag(server);
    del_server(&servFdTree, server, BY_FD);
    *fd = server->fd = -1;
    if (server->connecting > 0)
	server->connecting = 0;
    if (server_tool)
	fl_trigger_object(server_tool->update);
    RemoveCommands(server);
    RemoveChannels(server);
}
static gChannel *
    get_nextgchannel(const Server * s)
{
/* 
 * search global channel list for  servername */
    gChannel *p = chanStart;

    if (!s)
	return NULL;
    while (p != chanEnd)
	if (
	       !strcasecmp(p->server, s->servname) &&
	       s->port == p->port
	   )
	    return p;
	else
	    p = p->next;
    return NULL;
}
static void RemoveChannels(Server * s)
{
    gChannel *p;
    short i = 0;

    assert(s);
    while ((p = get_nextgchannel(s))) {
	gChannel *q;

	free(p->name);
	free(p->server);
	q = p->next;
	if (q == chanEnd)
	    chanEnd = p;
	else
	    *p = *q;
	free(q);
	i++;
    }
}
static void RemoveCommands(Server * s)
{
/* remove all commands sent to this server */
    CmdSent *p = cmdsentStart, *q;

    while (p != cmdsentEnd) {
	if (p->win->server == s) {
	    q = p->next;
	    free(p->cmd);
	    free(p->arg);
	    if (q == cmdsentEnd)
		cmdsentEnd = p;
	    else
		*p = *q;
	    free(q);
	}
	p = p->next;
    }
}

int sendto_server(Winstruct * win, const char *buf)
{
    assert(win && buf);
    if (!win->server || win->server->fd < 0) {// dont even think about rming this line
	fl_set_object_label(win->chanwin->WinStatusLine,
			    "You are not connected to a server");
	fl_ringbell(0);
	return -1;
    } else{
#if 0
    /*disenabled so as to support named server connections */
     if (win->server->connecting > 0) {
	add_cmdoncon(SERV_CMD, (char *) buf,
		     win->server->servname, win->server->port);
	return 0;
    }
#endif
    int n;
    n = strlen(buf);
    if (writen(win->server->fd, buf, n) != n) {
	if (errno != EAGAIN) {
	    say2(-1, 1, "Write error to port %d of %s: %s",
		 win->server->port, win->server->name,strerror(errno));
	    fl_ringbell(0);
	    close_server_connection(&win->server->fd);
	    return -2;
	}
    }
    win->server->wrote += n;
    }
    return 0;
}

void close_server_connectionbyname(char *servname, int port)
{
    Server *server;

    server = search_server_tree_byname(servname, port);

    if (!server)
	error(0, "%s: No connection to port %d found", servname, port);
    else if (server->fd < 0)
	error(0, "You are not connected to port %d of %s",
	      server->port, server->servname, "");
    else {
	WinList *p;

	build_winlistbyfd(server->fd);
	p = servWin;
	while (p) {
	    ClearWinChanList(p->win);
	    UpdateStatus(p->win);
	    p = p->next;
	}
	new_close_connection(&server->fd, NULL, 1);
    }
}

static char *
    get_serverline(FILE * fp)
{
/* 
 * read one line from serverlist  */
    char p[513];
    char *buf;

    while (fgets(p, 512, fp)) {
	p[strlen(p) - 1] = '\0';
	if (!(buf = clean(p)) || *buf == COMMENT || *buf == SERV_DELIM_CHR)
	    continue;
	else
	    return buf;
    }
    return NULL;
}

static int get_server(FILE * fp, Server * server)
{
/* 
 * get line of form  server:port:passwd:nick:email:ircname return <0: EOF
 * return >1:item format error.. try next item.
 * 
 */
    char *buf;
    char *tmp;			/*password, *nick, *ircname, *email; */
    size_t ll;
    int i;
    char *ptr;

    if ((buf = get_serverline(fp)) == NULL)
	return EOF;
    ll = strlen(buf);

    tmp = (char *) malloc(sizeof(char) * ll);

    if (!tmp) {
	error(1, "malloc error");
	return errno;
    }
    ptr = tmp;

    while (*buf && *buf != ':')
	*ptr++ = *buf++;
    if (*buf == ':') {
	*ptr = '\0';
	buf++;
    }
    if (!(*tmp))
	return -2;
    else
	server->servname = strdup(tmp);
    *tmp = 0;
    ptr = tmp;
    while (*buf && *buf != ':')
	*ptr++ = *buf++;
    if (*buf == ':') {
	*ptr = '\0';
	buf++;
    }
    i = atoi(tmp);
    if (i < 0)
	*tmp = '\0';
    ll = strlen(tmp);
    while ((i = atoi(tmp)) > USHRT_MAX)
	*(tmp + --ll) = '\0';
    if ((server->port = i) == 0)
	server->port = DEFAULT_PORT;
    *tmp = 0;
    ptr = tmp;
    while (*buf && *buf != ':')
	*ptr++ = *buf++;
    if (*buf == ':') {
	*ptr = '\0';
	buf++;
    }
    if (!(*tmp)) {
	server->password = NULL;
    }
    else
	server->password = strdup(tmp);
    *tmp = 0;
    ptr = tmp;
    while (*buf && *buf != ':')
	*ptr++ = *buf++;
    if (*buf == ':') {
	*ptr = '\0';
	buf++;
    }
    if (!(*tmp))
	server->nick = strdup(Name);
    else
	server->nick = strdup(tmp);
    *tmp = 0;
    ptr = tmp;
    while (*buf && *buf != ':')
	*ptr++ = *buf++;
    if (*buf == ':') {
	*ptr = '\0';
	buf++;
    }
    if (!(*tmp)) {
	server->email = (char *) malloc(sizeof(char) * (strlen(Name) + strlen(host) + 3));

	sprintf(server->email, "%s@%s", Name, host);
    }
    else
	server->email = strdup(tmp);
    *tmp = 0;
    ptr = tmp;
    while ((*ptr++ = *buf++));
    if (!(*tmp)) {
	server->ircname = (char *) malloc(sizeof(char) * 
	(strlen(sula_NAME sula_VERSION) + 8));

	strcpy(server->ircname, "Using " sula_NAME sula_VERSION);
    }
    else
	server->ircname = strdup(tmp);
    free(tmp);
    server->mode = NULL;
    server->flag = SERVER_FLAG_RESET;
    server->state = -1;
    server->fd = -1;
    server->lastInvite = NULL;
    server->name = NULL;
    server->left = server->right = NULL;
    server->iNick = get_list_bylabel("default");
    if (server->iNick < 0) {
	server->iNick = new_nicklist("default");
	if (server->iNick < 0)
	    error(-2, "[%s %d] Error", __FILE__, __LINE__);
    }
    aNickList[server->iNick].current =
       add_nick_to_list(server->iNick, server->nick);
    return 0;
}
